/*
 * client_config.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "control.hpp"
#include "client_config.hpp"
#include "gfx_manager.hpp"
#include "misc/my_exceptions.hpp"
#include "potion_renderer.hpp"
#include "skull_renderer.hpp"

ClientConfig::ClientConfig(Coercri::NetworkConnection &remote, GfxManager &gm)
{
    // NOTE: we load all the gfx config, overlay offsets etc from the network. This probably
    // means that config needs to be the same on both ends (eg changing the gfx sizes would
    // mess up the overlay offsets) ... Coming up with a better method needs some thought
    // though.
    
    InitOverlayOffsets(remote);
    
    int n;
    n = remote.getLong();
    graphics.resize(n);
    for (int i=0; i<n; ++i) {
        graphics[i] = gm.createGraphicFromNetwork(remote);
    }

    n = remote.getLong();
    anims.resize(n);
    for (int i=0; i<n; ++i) {
        anims[i] = new Anim(remote, *this);
    }

    n = remote.getLong();
    overlays.resize(n);
    for (int i=0; i<n; ++i) {
        overlays[i] = new Overlay(remote, *this);
    }

    n = remote.getLong();
    if (n != NUM_STANDARD_CONTROLS) throw UnexpectedError("standard controls do not match");
    n = remote.getLong();
    controls.resize(n+NUM_STANDARD_CONTROLS);
    for (int i=0; i<NUM_STANDARD_CONTROLS; ++i) {
        controls[i] = g_standard_controls[i];
    }
    for (int i=NUM_STANDARD_CONTROLS; i<n+NUM_STANDARD_CONTROLS; ++i) {
        controls[i] = new Control(remote, *this, i+1);
    }

    potion_renderer.reset(new PotionRenderer(remote, *this));
    skull_renderer.reset(new SkullRenderer(remote, *this));

    menu_centre = getGraphic(remote.getLong());
    menu_empty = getGraphic(remote.getLong());
    menu_highlight = getGraphic(remote.getLong());
}

ClientConfig::~ClientConfig()
{
    // NOTE: do not delete controls 0 to NUM_STANDARD_CONTROLS-1 as these point to g_standard_controls
    // (i.e. they were not created with new!)
    for (int i=NUM_STANDARD_CONTROLS; i<controls.size(); ++i) delete controls[i];
    for (int i=0; i<overlays.size(); ++i) delete overlays[i];
    for (int i=0; i<anims.size(); ++i) delete anims[i];
}

Coercri::Graphic * ClientConfig::getGraphic(int id) const
{
    if (id < 1 || id > graphics.size()) return 0;
    else return graphics[id-1];
}

Anim * ClientConfig::getAnim(int id) const
{
    if (id < 1 || id > anims.size()) return 0;
    else return anims[id-1];
}

Overlay * ClientConfig::getOverlay(int id) const
{
    if (id < 1 || id > overlays.size()) return 0;
    else return overlays[id-1];
}

Control * ClientConfig::getControl(int id) const
{
    if (id < 1 || id > controls.size()) return 0;
    else return controls[id-1];
}

