/*
 * FILE:
 *   enet_network_driver.cpp
 *
 * AUTHOR:
 *   Stephen Thompson, 2008
 *
 * COPYRIGHT:
 *   Public Domain
 *
 */

#include "enet_network_connection.hpp"
#include "enet_network_driver.hpp"

#include "../core/coercri_error.hpp"

namespace Coercri {

    namespace {
        template<class T>
        struct PtrEq {
            PtrEq(T *p_) : p(p_) { }
            bool operator()(const boost::shared_ptr<T> &q) const { return q.get() == p; }
            T* p;
        };
    }

    bool EnetNetworkDriver::is_enet_initialized = false;
    
    EnetNetworkDriver::EnetNetworkDriver(int max_inc, int max_outgoing)
        : incoming_host(0), outgoing_host(0), server_port(-1), server_enabled(true), max_incoming(max_inc)
    {
        if (is_enet_initialized) {
            throw CoercriError("enet initialized twice");
        }

        if (enet_initialize() != 0) {
            throw CoercriError("enet initialization failed");
        }
        
        is_enet_initialized = true;

        // create the outgoing_host. (incoming_host only created as needed.)
        outgoing_host = enet_host_create(0, max_outgoing, 0, 0);
        if (!outgoing_host) {
            throw CoercriError("EnetNetworkDriver: enet_host_create failed");
        }
    }

    EnetNetworkDriver::~EnetNetworkDriver()
    {
        if (outgoing_host) enet_host_destroy(outgoing_host);
        if (incoming_host) enet_host_destroy(incoming_host);
        enet_deinitialize();
        is_enet_initialized = false;
    }

    boost::shared_ptr<NetworkConnection> EnetNetworkDriver::openConnection(const std::string &hostname, int port_number)
    {
        boost::shared_ptr<EnetNetworkConnection> new_conn(new EnetNetworkConnection(outgoing_host, hostname, port_number));
        connections_out.push_back(new_conn);
        return new_conn;
    }

    void EnetNetworkDriver::setServerPort(int port)
    {
        if (port != server_port && incoming_host) {
            // Disconnect all current incoming connections
            for (EnetConnections::iterator it = connections_in.begin(); it != connections_in.end(); ++it) {
                (*it)->onDisconnect();
            }
            connections_in.clear();
            new_connections_in.clear();

            // Destroy the host. (This will automatically reset all connections.)
            destroyIncomingHost();
        }

        server_port = port;
    }

    void EnetNetworkDriver::createIncomingHostIfNeeded()
    {
        if (!incoming_host && server_enabled && server_port >= 0) {
            // Create the incoming host
            ENetAddress address;
            address.host = ENET_HOST_ANY;
            address.port = server_port;
            incoming_host = enet_host_create(&address, max_incoming, 0, 0);
            if (!incoming_host) {
                throw CoercriError("EnetNetworkDriver: enet_host_create failed");
            }
        }
    }

    void EnetNetworkDriver::destroyIncomingHost()
    {
        if (incoming_host) {
            enet_host_destroy(incoming_host);
            incoming_host = 0;
        }
    }
    
    void EnetNetworkDriver::enableServer(bool enabled)
    {
        server_enabled = enabled;
        if (!server_enabled && incoming_host) {
            // Close any existing incoming connections
            for (EnetConnections::iterator it = connections_in.begin(); it != connections_in.end(); ++it) {
                (*it)->close();
            }
        }
    }
    
    EnetNetworkDriver::Connections EnetNetworkDriver::pollIncomingConnections()
    {
        Connections result;
        std::swap(new_connections_in, result);
        return result;
    }

    bool EnetNetworkDriver::serviceHost(ENetHost *host)
    {
        if (!host) return false;

        ENetEvent event;
        int result = enet_host_service(host, &event, 0);
        
        if (result < 0) {
            throw CoercriError("EnetNetworkDriver: enet_host_service failed");
        } else if (result == 0) {
            return false;
        } else {
            // got an event
            switch (event.type) {
            case ENET_EVENT_TYPE_CONNECT:
                if (event.peer->data) {
                    // One of my outgoing connections has been acknowledged by the remote server.
                    EnetNetworkConnection *conn = static_cast<EnetNetworkConnection*>(event.peer->data);
                    conn->onReceiveAcknowledgment();
                } else {
                    // This is a new incoming connection.
                    if (server_enabled) {
                        boost::shared_ptr<EnetNetworkConnection> new_conn(new EnetNetworkConnection(event.peer));
                        connections_in.push_back(new_conn);
                        new_connections_in.push_back(new_conn);
                    } else {
                        // refuse the connection
                        enet_peer_disconnect(event.peer, 0);
                    }
                }
                break;
                
            case ENET_EVENT_TYPE_DISCONNECT:
                // We've been disconnected. Note that we get this event for both timeouts and explicit disconnects.
                {
                    EnetNetworkConnection *conn = static_cast<EnetNetworkConnection*>(event.peer->data);
                    conn->onDisconnect();  // Goes into disconnected state, and breaks link to ENetPeer object.
                    
                    // Remove the connection from my vectors
                    connections_out.erase(std::remove_if(connections_out.begin(), connections_out.end(),
                                                         PtrEq<EnetNetworkConnection>(conn)),
                                          connections_out.end());
                    connections_in.erase(std::remove_if(connections_in.begin(), connections_in.end(),
                                                        PtrEq<EnetNetworkConnection>(conn)),
                                         connections_in.end());
                    new_connections_in.erase(std::remove_if(new_connections_in.begin(), new_connections_in.end(),
                                                            PtrEq<NetworkConnection>(conn)),
                                             new_connections_in.end());

                    // If server is disabled and there are no incoming connections left, then we may
                    // destroy the server host
                    if (connections_in.empty() && !server_enabled) {
                        destroyIncomingHost();
                    }
                }
                break;
                
            case ENET_EVENT_TYPE_RECEIVE:
                // A packet has been received.
                {
                    EnetNetworkConnection* from_conn = static_cast<EnetNetworkConnection*>(event.peer->data);
                    from_conn->onReceivePacket(event.packet);
                }
                break;
            }
            return true;
        }
    }

    bool EnetNetworkDriver::doEvents()
    {
        createIncomingHostIfNeeded();
        
        bool did_something = false;
        did_something = serviceHost(outgoing_host) || did_something;
        did_something = serviceHost(incoming_host) || did_something;
        return did_something;
    }
}
