/*
 * FILE:
 *   network_connection.hpp
 *
 * PURPOSE:
 *   Represents a network connection to a foreign host. Can send and
 *   receive packets. Note: data may not be actually sent/received
 *   until NetworkDriver::pollEvents is called.
 *
 * AUTHOR:
 *   Stephen Thompson, 2008
 *
 * COPYRIGHT:
 *   Public Domain
 *
 */

#ifndef COERCRI_NETWORK_CONNECTION_HPP
#define COERCRI_NETWORK_CONNECTION_HPP

#include <string>
#include <vector>

namespace Coercri {

    class NetworkConnection {
    public:
        NetworkConnection() : incoming_index(0) { }
        virtual ~NetworkConnection() { }

        // Possible connection states
        enum State {
            PENDING,    // (Client side only) Awaiting connection to server
            CONNECTED,  // Ready to send/receive data
            CLOSED      // Connection closed (either by local or remote close() call, or by a timeout).
        };
        
        // Query functions
        // NOTE: a CLOSED or FAILED connection may still be able to read data (if there has been buffering).
        // In particular it is always possible to read to the end of the current packet.
        virtual State getState() const = 0;
        bool inAvail();    // True if at least one byte can be read.

        // Control functions
        virtual void close() = 0;
        
        // Read functions
        unsigned char getUbyte();
        signed char getByte();
        unsigned short int getUshort();
        short int getShort();
        unsigned int getUlong();
        int getLong();
        int getCInt();   // "compressed integer", can store values between -268435456 and +268435455
        std::string getString();
        
        // Write functions
        void putUbyte(unsigned char);
        void putByte(signed char);
        void putUshort(unsigned short int);
        void putShort(short int);
        void putUlong(unsigned int);
        void putLong(int);
        void putCInt(int);
        void putString(const std::string &);
        void endPacket();  // Ends the current packet (nothing will be sent until this is called)
        
    protected:
        typedef unsigned char ubyte;

        // Send a packet (represented as a vector<ubyte>)
        virtual void transmit(const std::vector<ubyte> &) = 0;

        // Receive a packet into buf if there is one. (Do nothing otherwise.)
        // Precondition: buf is empty.
        // (This function should not block. Presumably subclasses will do some sort of buffering
        // based on network events received in NetworkDriver::pollEvents().)
        virtual void receive(std::vector<ubyte> &buf) = 0;

    private:
        std::vector<ubyte> outgoing_packet;
        std::vector<ubyte> incoming_packet;
        int incoming_index;  // position within the incoming packet
    };
}

#endif
