/*
 * connecting_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "connecting_screen.hpp"
#include "knights_game.hpp"
#include "loading_screen.hpp"
#include "title_screen.hpp"
#include "x_centre.hpp"

#include "gui_button.hpp"
#include "gui_centre.hpp"
#include "gui_panel.hpp"

#include "coercri/gfx/font.hpp"
#include "coercri/gfx/gfx_context.hpp"
#include "coercri/gfx/color.hpp"
#include "coercri/gfx/window_listener.hpp"

#include "boost/scoped_ptr.hpp"

class ConnectingScreenImpl : public gcn::ActionListener {
public:
    ConnectingScreenImpl(KnightsGame &kg, gcn::Gui &gui, shared_ptr<Coercri::Window> win);
    void action(const gcn::ActionEvent &event);

    KnightsGame & getKnightsGame() const { return knights_game; }
    bool awaitingInput() const { return awaiting_input; }
    const char * getIPAddress() const { return address_field->getText().c_str(); }
private:
    KnightsGame &knights_game;
    shared_ptr<Coercri::Window> window;
    
    scoped_ptr<GuiCentre> centre;
    scoped_ptr<GuiPanel> panel;
    scoped_ptr<gcn::Container> container;
    scoped_ptr<gcn::Label> label;
    scoped_ptr<gcn::TextField> address_field;
    scoped_ptr<gcn::Button> cancel_button, connect_button;

    bool awaiting_input;
};

ConnectingScreenImpl::ConnectingScreenImpl(KnightsGame &kg, gcn::Gui &gui, shared_ptr<Coercri::Window> win)
    : knights_game(kg), window(win)
{
    awaiting_input = !kg.server;
    
    container.reset(new gcn::Container);
    container->setOpaque(false);
    
    const int pad = 10;
    int y = pad;
    
    if (knights_game.server) {
        label.reset(new gcn::Label("Waiting for connection"));
    } else {
        label.reset(new gcn::Label("Enter IP address to connect to:"));
    }
    container->add(label.get(), pad, y);
    y += label->getHeight() + pad;
    
    if (!knights_game.server) {
        address_field.reset(new gcn::TextField);
        address_field->adjustSize();
        address_field->setWidth(label->getWidth());
        address_field->addActionListener(this);
        container->add(address_field.get(), pad, y);
        y += address_field->getHeight() + pad;
    }
    
    cancel_button.reset(new GuiButton("Cancel"));
    cancel_button->addActionListener(this);
    if (knights_game.server) {
        container->add(cancel_button.get(), pad + label->getWidth() / 2 - cancel_button->getWidth() / 2, y);
    } else {
        connect_button.reset(new GuiButton("Connect"));
        connect_button->addActionListener(this);
        container->add(connect_button.get(), pad, y);
        container->add(cancel_button.get(), pad + label->getWidth() - cancel_button->getWidth(), y);
    }
    container->setSize(2*pad + label->getWidth(), y + cancel_button->getHeight() + pad);
    
    panel.reset(new GuiPanel(container.get()));
    centre.reset(new GuiCentre(panel.get()));
    gui.setTop(centre.get());

    // make sure the text field is focused initially
    if (!knights_game.server) address_field->requestFocus();
}

void ConnectingScreenImpl::action(const gcn::ActionEvent &event)
{
    if (event.getSource() == connect_button.get() || event.getSource() == address_field.get()) {
        awaiting_input = false;
        connect_button->setVisible(false);
        cancel_button->setX(panel->getWidth() / 2 - cancel_button->getWidth() / 2);
        address_field->setVisible(false);
        label->setCaption("Connecting to " + address_field->getText() + "...");
        label->setAlignment(gcn::Graphics::CENTER);
        window->invalidateAll();
        knights_game.repeatLastMouseInput();

        // Create the network connection
        knights_game.remote = knights_game.net_driver->openConnection(address_field->getText(),
                                                                      CfgInt("port_number"));
    } else if (event.getSource() == cancel_button.get()) {
        // Go back to title screen
        auto_ptr<Screen> title_screen(new TitleScreen);
        knights_game.requestScreenChange(title_screen);
    }
}

bool ConnectingScreen::start(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &gui)
{
    pimpl.reset(new ConnectingScreenImpl(kg, gui, win));
    
    // Enable server if required
    if (kg.server) {
        kg.net_driver->enableServer(true);
    }

    return true;
}

void ConnectingScreen::update()
{
    bool done = false;

    if (pimpl->getKnightsGame().server) {
        // Listen for clients connecting to the server
        Coercri::NetworkDriver::Connections new_connections = pimpl->getKnightsGame().net_driver->pollIncomingConnections();
        if (!new_connections.empty()) {
            // Somebody has connected -- we are done.
            ASSERT(new_connections.size() == 1);
            pimpl->getKnightsGame().remote = new_connections.front();
            done = true;
        }
    } else {
        if (!pimpl->awaitingInput()) {
            Coercri::NetworkConnection::State state = pimpl->getKnightsGame().remote->getState();
            if (state == Coercri::NetworkConnection::CLOSED) {
                // Our connection timed out, or was refused. Go back to title screen.
                auto_ptr<Screen> title_screen(new TitleScreen);
                pimpl->getKnightsGame().requestScreenChange(title_screen);
            } else if (state == Coercri::NetworkConnection::CONNECTED) {
                // Let the game begin!
                done = true;
            }
            // Otherwise -- we are in PENDING state. Keep waiting for the connection.
        }
    }

    if (done) {
        // Go to the loading screen next.
        auto_ptr<Screen> loading_screen(new LoadingScreen);
        pimpl->getKnightsGame().requestScreenChange(loading_screen);
    }
}


