/*
 * dungeon_view.hpp
 *
 * Receives updates from Game about what a given player can see in the
 * dungeon.
 *
 * Also stores information about which squares have currently been
 * mapped. (This is so that we can program Sense Items -- which
 * highlights items on the mini-map, but only for squares which your
 * knight has already explored.)
 * 
 * Also stores Controller information.
 * 
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef DUNGEON_VIEW_HPP
#define DUNGEON_VIEW_HPP

#include "map_support.hpp"
#include "mini_map_colour.hpp"
#include "potion_magic.hpp"

#include "coercri/gfx/graphic.hpp"

#include "boost/shared_ptr.hpp"
#include "boost/weak_ptr.hpp"
using namespace boost;

#include <string>
#include <vector>
using namespace std;

class Anim;
class ColourChange;
class Control;
class Entity;
class Overlay;

class DungeonView {
public:
    virtual ~DungeonView() { }
    
    //
    // Room control:
    // The DungeonView remembers tiles and items from different rooms, so when the player 
    // revisits a previous room, the tiles and items don't have to be downloaded again.
    // Entities are never remembered though -- the ViewManager has to resend all entities
    // (using addEntity) after calling setCurrentRoom.
    //
    // NB: setCurrentRoom(-1) means display a blank screen (eg if the player has died). 
    //

    virtual void setCurrentRoom(int r, int width, int height) = 0;  // move the display to a new room. (Removes all entities)
    virtual int getCurrentRoom() const = 0;  // which room is currently being displayed? (returns -1 for "none")
    virtual void mapCurrentRoom(const MapCoord &tl_pos) = 0;  // place room onto mini-map
    virtual void mapWall(const MapCoord &mc) = 0;  // place a COL_WALL square onto mini-map
    
    // mapKnightLocation: show a knight on the mini-map. n is the player number (from
    // 0 to NPLYRS-1); mc is the position, or null if knight is not to be shown.
    virtual void mapKnightLocation(int n, const MapCoord &mc) = 0;
    
    // mapItemLocation: switch on or off an "item indicator" on the mini-map at a given
    // location. (used for sense items.)
    virtual void mapItemLocation(const MapCoord &mc, bool on) = 0;

    // Wipe Map
    virtual void wipeMap() = 0;

    // Read Map
    virtual bool isSquareMapped(const MapCoord &mc) const = 0;
    

    //
    // Dungeon updates:
    // (These all apply to the current room only. Coords are given relative to bottom left
    // of current room.)
    //

    // entities

    // NOTE: all entities are assigned an "id" number by
    // ViewManager. id==0 is reserved for "my knight".
    
    // NOTE for setAnimData: if atz_diff==0 this means anim_tzero == 0
    // and if atz_diff>0 this means anim_tzero == atz_diff + GVT.
    
    virtual void addEntity(unsigned short int id, int x, int y, MapHeight ht, MapDirection facing,
                           const Anim * anim, const Overlay *ovr, int af, int atz_diff,
                           bool ainvuln, // (anim data)
                           int cur_ofs, MotionType motion_type, int motion_time_remaining) = 0;
    virtual void rmEntity(unsigned short int id) = 0;
    virtual void repositionEntity(unsigned short int id, int new_x, int new_y) = 0;
    virtual void moveEntity(unsigned short int id, MotionType motion_type,
                            int motion_duration, bool missile_mode) = 0;
    virtual void flipEntityMotion(unsigned short int id, int initial_delay, int motion_duration) = 0;
    virtual void setAnimData(unsigned short int id, const Anim *, const Overlay *, int af,
                             int atz_diff, bool ainvuln, bool currently_moving) = 0;
    virtual void setFacing(unsigned short int id, MapDirection new_facing) = 0;

    // tiles
    virtual void clearTiles(int x, int y) = 0;  // wipe all tiles at a coord.
    virtual void setTile(int x, int y, int depth, const Coercri::Graphic * gfx,
                         shared_ptr<const ColourChange> cc, MiniMapColour col) = 0;

    // items
    virtual void setItem(int x, int y, const Coercri::Graphic * gfx) = 0;

    // "icons" (used for blood splats)
    virtual void placeIcon(int x, int y, const Coercri::Graphic *g, int dur) = 0;

    
    //
    // UI / Status area displays:
    //

    // backpack control.
    // Game tells dungeonviews about slot number, graphic to display in that slot
    // (from item_type->backpack_graphic), how many the player is carrying, and
    // the maximum that the player is allowed to carry (or 0 for no maximum).
    // Slot conventions:
    // 101,102,... normal items
    // 200 for the lock picks, and 201,202,... for keys (one max of each)
    virtual void setBackpack(int slot, const Coercri::Graphic *gfx, const Coercri::Graphic *overdraw, int no_carried, int no_max) = 0;

    // add a death (skull)
    virtual void addSkull() = 0;

    // set health and magics
    virtual void setHealth(int h) = 0;
    virtual void setPotionMagic(PotionMagic pm, bool poison_immunity) = 0;
    
    // display a message eg "QUICKNESS"
    virtual void flashMessage(const string &msg, int ntimes) = 0;

    // continuous messages (paralyzation; hint text when approaching exit)
    virtual void cancelContinuousMessages() = 0;
    virtual void addContinuousMessage(const string &msg) = 0;

    // flash the screen (as after stepping on a pentagram, say)
    virtual void flashScreen(int delay) = 0;

    // sound
    // NOTE: flag is set to true if the sound was actually played to
    // the speakers. this is a hack to prevent the same sound being
    // played more than once on the local machine (in split screen
    // mode) -- necessary because otherwise the sound will play twice
    // as loudly as it should when two people are in the same room.
    virtual void playSound(const string &name, int frequency, bool &flag) = 0;
    
    
    //
    // winning and losing
    //
    virtual void winGame() = 0;
    virtual void loseGame() = 0;
    virtual void quitGame() = 0;
    
    
    //
    // Controller status.
    //

    // set the (currently) available knight-actions and the highlighted menu option.
    // The "bool" in the pair<const Control*, bool> indicates whether this is a "primary"
    // (TRUE) or "secondary" (FALSE) controls. (Secondary controls are only accessible
    // through the menu, and cannot be accessed by tapping fire.)
    virtual void setAvailableControls(const vector<pair<const Control *,bool> >
                                      &available_controls) = 0;
    virtual void setMenuHighlight(const Control *ctrl) = 0;

    // readController(): get current controller state.
    // -- This is called if the knight is waiting to do something.
    // -- It should return either NULL (knight doesn't want to do
    // anything), or one of the controls in "available_controls" (see
    // setAvailableControls), or one of the "standard" controls that
    // are considered to be always available (move, attack, or
    // withdraw; see control.hpp).
    virtual const Control * readController() = 0;

    // Update function. This should be called 'periodically' (once per game update, say).
    // (This is where RemoteDungeonView flushes its output buffer -- LocalDungeonView does
    // not use this.)
    virtual void update() { }


    // Cheat controls (Optional)
    // Note that the vector is not copied.
    virtual void addCheatControls(const vector<const Control*> *controls) { }
};

#endif

