/*
 * gui_text_wrap.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "gui_text_wrap.hpp"

#include <cctype>

struct GTWPrinter {
    virtual void operator()(const std::string &text, int y, bool do_centre) = 0;
};

namespace {
    const int pad = 3;
    
    struct ActualPrinter : GTWPrinter {
        ActualPrinter(int w, gcn::Graphics *g) : width(w), graphics(g) { }
        
        void operator()(const std::string &text, int y, bool do_centre)
        {
            if (do_centre) {
                graphics->drawText(text, width/2, y, gcn::Graphics::CENTER);
            } else {
                graphics->drawText(text, pad, y);
            }
        }

        int width;
        gcn::Graphics *graphics;
    };
}

GuiTextWrap::GuiTextWrap()
 : opaque(false), centred(false), rich(false)
{
    addMouseListener(this);
}

void GuiTextWrap::draw(gcn::Graphics *graphics)
{
    if (isOpaque()) {
        // Clear the background first.
        graphics->setColor(getBackgroundColor());
        graphics->fillRectangle(gcn::Rectangle(0, 0, getWidth(), getHeight()));
    }
    
    graphics->setFont(getFont());

    ActualPrinter p(getWidth(), graphics);
    doDraw(&p);
}

void GuiTextWrap::adjustHeight()
{
    setHeight(doDraw(0));
}

int GuiTextWrap::doDraw(GTWPrinter *printer)
{
    const int max_line_width = getWidth() - 2*pad;    
    gcn::Font *font = getFont();

    bool do_centre = isCentred();
    std::string line;
    int y = 0;
    std::string::size_type fit_chars = 0;
    std::string::const_iterator it = text.begin();
    
    while (1) {
        char c = 0;
        bool at_newline = false;
        if (it == text.end()) {
            if (!line.empty()) {
                // We've reached the end of input but there is still stuff in the buffer.
                // Manufacture an extra newline character, this should force it to print.
                at_newline = true;
            } else {
                // We've printed everything -- exit.
                break;
            }
        } else if (*it == '%' && isRich()) {
            // Intercept % as "rich text" marker
            ++it;
            if (it != text.end()) {
                if (*it == '%') {
                    // A normal % sign...
                    c = '%';
                } else if (*it == 'c') {
                    // Directive to centre text
                    do_centre = true;
                } else if (*it == 'l') {
                    // Directive to left-justify text
                    do_centre = false;
                }
                // (%-anything-else is just swallowed.)
                ++it;
            }
        } else if (*it == '\n') {
            at_newline = true;
            ++it;
        } else {
            // Normal character -- fetch it in, and increment the iterator.
            c = *it;
            ++it;
        }

        // Append the character to the line.
        if (c != 0) line += c;

        // Calculate the new line length
        const int line_width = font->getWidth(line);
        
        // Check to see if we are at a possible line break point.
        // (This means a non-space character where the next character IS a space.)
        if (c != 0 && !std::isspace(c) && (it == text.end() || std::isspace(*it))) {
            if (line_width <= max_line_width) {
                // OK, approve this point as a possible line break position.
                fit_chars = line.length();
            }
        }

        // If the line has got too big (or we have reached an explicit
        // newline) then we should fdump everything up to the previous
        // line break point, then start a new line.
        if (line_width > max_line_width || at_newline) {
            // Special case: if fit_chars == 0 (&& !line.empty()) then
            // reset fit_chars to line.length() [or line.length() - 1
            // if the entire line would be too big]. This is the case
            // where the entire line is too big to print, but there
            // were no possible line break points.
            if (fit_chars == 0 && !line.empty()) {
                if (at_newline) fit_chars = line.length();
                else fit_chars = line.length()-1;
            }
            
            // Print out the line (up to the last approved break point)
            const std::string to_print = line.substr(0, fit_chars);
            if (printer) (*printer)(to_print, y, do_centre);
            y += font->getHeight();

            // Reset for the next line (skip any spaces)
            while (fit_chars < line.length() && std::isspace(line[fit_chars])) ++fit_chars;
            line = line.substr(fit_chars, std::string::npos);
            fit_chars = 0;
        }
    }

    return y;
}

void GuiTextWrap::mouseReleased(gcn::MouseEvent &mouseEvent)
{
    distributeActionEvent();
}
