/*
 * in_game_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "client_config.hpp"
#include "game_client.hpp"
#include "game_server.hpp"
#include "gfx_manager.hpp"
#include "gfx_resizer.hpp"
#include "in_game_screen.hpp"
#include "knights_game.hpp"
#include "local_dungeon_view.hpp"
#include "mediator.hpp"
#include "menu_screen.hpp"
#include "misc/my_exceptions.hpp"
#include "remote_dungeon_view.hpp"
#include "server_config.hpp"
#include "task_manager.hpp"
#include "x_centre.hpp"

#include "coercri/gfx/rectangle.hpp"

unsigned int InGameScreen::getUpdateInterval()
{
    return 1000 / static_cast<unsigned int>(CfgInt("fps"));
}

InGameScreen::InGameScreen(KnightsGame &kg, const MenuSelections &msel)
    : knights_game(kg), first_update(false), sent_ready_yet(false), knights_spawned_yet(false),
      prev_update_time(0), stage(SETUP_STAGE)
{
    if (knights_game.server) {
        // Server setup is done here (client setup is done in separate "setup stage")
        serverSetup(msel);
        stage = HANDSHAKE_STAGE;
    }
}

InGameScreen::~InGameScreen()
{
    // Upon destruction of InGameScreen (end of game) we can delete the GameClient,
    // GameServer and the DungeonViews.
    game_client.reset();
    game_server.reset();
    rdv1.reset();
    ldv2.reset();
    ldv1.reset();

    // Remove myself from the window if needed
    if (window) window->rmWindowListener(this);
}

namespace {
    // Helper routines to configure DungeonViews properly.
    
    // Note there are two versions, one based on a ServerConfig and one that sends/receives
    // the necessary information over the network (to clients).

    // [Maybe potion and skull renderers, and menu gfx, should be shared? (TODO)]

    void SetupLocalDungeonView(LocalDungeonView &ldv, ServerConfig &server_config)
    {
        ldv.setMiniMap(Mediator::instance().getMap()->getWidth(),
                       Mediator::instance().getMap()->getHeight());
        ldv.setPotionRenderer(&server_config.getPotionRenderer());
        ldv.setSkullRenderer(&server_config.getSkullRenderer());
        ldv.setMenuGfx(server_config.getMenuCentre(), server_config.getMenuEmpty(),
                       server_config.getMenuHighlight());
    }
    
    void SendRemoteDviewSetup(Coercri::NetworkConnection &remote, GameServer &game_server)
    {
        remote.putCInt(Mediator::instance().getMap()->getWidth());
        remote.putCInt(Mediator::instance().getMap()->getHeight());
    }
    
    void ReceiveRemoteDviewSetup(LocalDungeonView &dview, Coercri::NetworkConnection &remote,
                                 ClientConfig &client_config)
    {
        int w = remote.getCInt();
        int h = remote.getCInt();
        dview.setMiniMap(w, h);
        dview.setPotionRenderer(client_config.getPotionRenderer());
        dview.setSkullRenderer(client_config.getSkullRenderer());
        dview.setMenuGfx(client_config.getMenuCentre(), client_config.getMenuEmpty(),
                         client_config.getMenuHighlight());
    }
}

void InGameScreen::serverSetup(const MenuSelections &msel)
{
    // First: create DungeonViews
    if (knights_game.remote) {
        // Server game
        ASSERT(knights_game.server_config);
        ldv1.reset(new LocalDungeonView(knights_game.getLeftController(),
                                        &knights_game.getRightController(),
                                        knights_game.getSoundManager()));
        rdv1.reset(new RemoteDungeonView(knights_game.remote, *knights_game.server_config, knights_game.getGfxManager()));
    } else {
        // Local game
        ldv1.reset(new LocalDungeonView(knights_game.getLeftController(), 0, knights_game.getSoundManager()));
        ldv2.reset(new LocalDungeonView(knights_game.getRightController(), 0, knights_game.getSoundManager()));
    }

    // Create GameServer
    vector<DungeonView*> dviews;
    if (ldv1) dviews.push_back(ldv1.get());
    if (ldv2) dviews.push_back(ldv2.get());
    if (rdv1) dviews.push_back(rdv1.get());
    game_server.reset(new GameServer(2));  // 2 players
    game_server->initialize(*knights_game.server_config, msel, dviews);

    // If we get to here, then the dungeon was successfully generated. Send out the 'M'
    // command to the remote (this closes the menu screen on the other side).
    if (knights_game.remote) {
        knights_game.remote->putUbyte('M');
        knights_game.remote->endPacket();
    }
    
    // Do local dview setup
    if (ldv1) SetupLocalDungeonView(*ldv1, *knights_game.server_config);
    if (ldv2) SetupLocalDungeonView(*ldv2, *knights_game.server_config);

    // Send out remote dview information
    if (knights_game.remote) {
        SendRemoteDviewSetup(*knights_game.remote, *game_server);
        knights_game.remote->endPacket();
    }
}

bool InGameScreen::tryClientSetup()
{
    // Create DungeonView if necessary
    if (!ldv1) {
        ldv1.reset(new LocalDungeonView(knights_game.getLeftController(),
                                        &knights_game.getRightController(),
                                        knights_game.getSoundManager()));
    }

    // Create GameClient if necessary
    if (!game_client) {
        // (note: GameClient ctor does not actually read anything from remote)
        game_client.reset(new GameClient(*knights_game.remote, *ldv1, *knights_game.client_config));
    }
    
    // Receive the remote dview information.
    // (This fails if there is no data available to read)
    if (!knights_game.remote->inAvail()) return false;
    ReceiveRemoteDviewSetup(*ldv1, *knights_game.remote, *knights_game.client_config);

    // We are done!
    return true;
}

void InGameScreen::onRepaint(Coercri::GfxContext &gc, const Coercri::Region &rgn)
{
    static const int ref_vp_width = CfgInt("dpy_viewport_width");
    static const int ref_vp_height = CfgInt("dpy_viewport_height");
    static const int ref_gutter = CfgInt("dpy_gutter");

    gc.clearClipRectangle();
    gc.clearScreen(Coercri::Color(0,0,0));
    if (!game_client && !game_server) {
        XCentre(gc, *knights_game.getFont(), knights_game.getFont()->getTextHeight() + 15, "STARTING GAME");
    } else {
        const int gvt = game_client? game_client->getGVT()
            : Mediator::instance().getGVT();

        // Have to compute appropriate scaled rectangles
        // NOTE: For split screen mode, assume 2 viewports, laid out horizontally.
        // For networked mode, we assume one viewport only.
        
        const int dpy_width = gc.getWidth();
        const int dpy_height = gc.getHeight();
        const int ref_width = ldv2 ? ref_vp_width * 2 + ref_gutter : ref_vp_width;
        const int ref_height = ref_vp_height;

        GfxManager &gm = knights_game.getGfxManager();
        float dummy;

        const float ideal_x_scale = float(dpy_width) / float(ref_width);
        const float ideal_y_scale = float(dpy_height) / float(ref_height);
        
        float x_scale, y_scale;
        gm.getGfxResizer()->roundScaleFactor(ideal_x_scale, x_scale, dummy);
        gm.getGfxResizer()->roundScaleFactor(ideal_y_scale, y_scale, dummy);
        const float scale = min(x_scale, y_scale);
        
        const int x_bias = int(ref_width * (ideal_x_scale - scale)) / 2;
        const int y_bias = int(ref_height * (ideal_y_scale - scale)) / 2;
        const int scaled_vp_width = int(ref_vp_width * scale);
        const int scaled_vp_height = int(ref_vp_height * scale);
        const int scaled_gutter = int(ref_gutter * scale);
        
        if (ldv1) {
            Coercri::Rectangle r(x_bias, y_bias, scaled_vp_width, scaled_vp_height);
            gc.setClipRectangle(r);
            ldv1->draw(gc, x_bias, y_bias, gm, scale, gvt);
        }
        if (ldv2) {
            const int left = x_bias + scaled_vp_width + scaled_gutter;
            Coercri::Rectangle r(left, y_bias, scaled_vp_width, scaled_vp_height);
            gc.setClipRectangle(r);
            ldv2->draw(gc, left, y_bias, gm, scale, gvt);
        }
        gc.clearClipRectangle();
    }
}

void InGameScreen::update()
{
    int c;
    
    switch (stage) {
    case SETUP_STAGE:
        ASSERT(!knights_game.server);
        if (tryClientSetup()) {
            stage = HANDSHAKE_STAGE;
        }
        break;
        
    case HANDSHAKE_STAGE:
        // At this point we do a crude synchronization between the two machines. This is
        // mainly to prevent the server from starting before the client is ready.
        
        if (!knights_game.remote) {
            // Don't need this for local games:
            stage = INGAME_STAGE;
            break;
        }
        
        // Send 'ready'
        if (!sent_ready_yet) {
            knights_game.remote->putUbyte('R');
            knights_game.remote->endPacket();
            sent_ready_yet = true;
        }
        
        // Wait for 'ready'
        if (!knights_game.remote->inAvail()) break;
        c = knights_game.remote->getUbyte();
        if (c != int('R')) {
            throw UnexpectedError("Handshake failure");
        }
        
        // OK, both sides are now roughly synchronized (up to lag delays etc).
        // Proceed to In-Game stage.
        stage = INGAME_STAGE;
        break;

    case INGAME_STAGE:
        {
            // Spawn knights if necessary
            if (!knights_spawned_yet && game_server) {
                game_server->spawnPlayers();
                knights_spawned_yet = true;
            }
            
            // Check for 'quit' commands on the client
            if (game_client && game_client->isGameQuit()) {
                stage = GAMEOVER_STAGE;
                break;
            }

            const int amount = getUpdateInterval() * CfgInt("ticks_per_second") / 1000;
            
            // Now do an actual update
             
            // NOTE: The in-game network protocol is actually quite simple:
            // (i) Server sends "dungeonview commands" to client (these are sent from RemoteDungeonView on
            // server side; on client side they are received in GameClient::update / GameClient::receiveCommands
            // and then forwarded to the client's LocalDungeonView).
            // (ii) Client sends "controls" to server - the controls are just 1 byte IDs. (These originate
            // in the client's LocalDungeonView, are transmitted from GameClient::update on client side,
            // and received in RemoteDungeonView::readController on server side.) Alternatively client may send
            // the byte 255 to acknowledge that the game has ended.
            
            if (game_client) {
                game_client->update(amount);
            } else if (game_server) {
                game_server->update(Mediator::instance().getGVT() + amount);
            }

            // force a repaint
            window->invalidateAll();
        }
        break;

    case GAMEOVER_STAGE:
        // If this is the server, we must now wait for the client to confirm that
        // it has received the "game over" message (no. 255; see also
        // RemoteDungeonView::readController and GameClient::sendGameOverMsg).
        if (game_server && knights_game.remote) {
            if (!knights_game.remote->inAvail()) break;
            c = knights_game.remote->getUbyte();
            if (c != 255) {
                // discard non-255 characters as there may still be incoming commands pending.
                break;
            }
        }

        // At this point we must go back to the menu screen.
        auto_ptr<Screen> screen(new MenuScreen);
        knights_game.requestScreenChange(screen);
        break;
    }
}

void InGameScreen::onMouseDown(int x, int y, Coercri::MouseButton b)
{
    if (b != Coercri::MB_LEFT) return;
    if (stage != INGAME_STAGE) return;
    if (game_client && game_client->isGameOver()
    || game_server && Mediator::instance().getWinningPlayer()) {
        stage = GAMEOVER_STAGE; // this will reset everything and go back to menu
    }
}

void InGameScreen::onRawKeyDown(Coercri::RawKey rk)
{
    if (game_server && stage == INGAME_STAGE) {
        const bool quit = (rk == Coercri::RK_ESCAPE);
        if (quit) {
            // tell server to send out QUIT messages to clients
            Mediator::instance().quitGame();
            // for our part, we go directly to GAMEOVER_STAGE...
            stage = GAMEOVER_STAGE;
        }
    }
}

bool InGameScreen::start(KnightsGame &, shared_ptr<Coercri::Window> win, gcn::Gui &)
{
    win->addWindowListener(this);
    window = win;
    return false;
}
