/*
 * menu_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "dungeon_generator.hpp"
#include "in_game_screen.hpp"
#include "knights_game.hpp"
#include "menu.hpp"
#include "menu_screen.hpp"
#include "menu_selections.hpp"
#include "my_exceptions.hpp"
#include "round.hpp"
#include "server_config.hpp"
#include "title_screen.hpp"
#include "x_centre.hpp"

#include "gui_button.hpp"
#include "gui_centre.hpp"
#include "gui_panel.hpp"

#include <sstream>
using namespace std;

// Undefine some annoying windows macros
#ifdef RGB
#undef RGB
#endif
#ifdef MB_LEFT
#undef MB_LEFT
#endif
#ifdef MB_RIGHT
#undef MB_RIGHT
#endif

namespace {
    const int extra_spacing = 2;   // extra spacing between menu lines (in pixels).

    MenuSelections g_menu_selections;
    bool g_first_time = true;
}

namespace {

    string MenuValueToString(int val, const MenuValue &mval)
    {
        if (!mval.getText().empty()) {
            return mval.getText();
        } else {
            // Assume that we want it drawn as a number
            // (Note: ignoring char-icon and graphic-icon settings for now)
            ostringstream oss;
            oss << val;
            return oss.str();
        }
    }
    
    class MenuListModel : public gcn::ListModel {
    public:
        virtual int getNumberOfElements() { return elts.size(); }
        virtual string getElementAt(int i) { if (i >= 0 && i < elts.size()) return elts[i]; else return ""; }
        vector<string> elts;
        vector<int> vals;
    };
    
    class MenuWidgets {
    public:
        MenuWidgets(const MenuItem &menu_item, gcn::ActionListener *listener)
        {
            label.reset(new gcn::Label(menu_item.getTitle()));
            label->adjustSize();
            list_model.reset(new MenuListModel);
            dropdown.reset(new gcn::DropDown(list_model.get()));
            dropdown->adjustHeight();
            dropdown->addActionListener(listener);
        }

        // NOTE: these have to be shared_ptrs since the MenuWidgets object gets copied.
        shared_ptr<gcn::Label> label;
        shared_ptr<gcn::DropDown> dropdown;
        shared_ptr<MenuListModel> list_model;
    };

    void ApplyMenuConstraints(const Menu &menu, MenuSelections &menu_selections)
    {
        // limit of 100 iterations to prevent infinite loops
        for (int count = 0; count < 100; ++count) {
            bool changed = false;
            for (int i=0; i<menu.getNumItems(); ++i) {
                const MenuItem &item(menu.getItem(i));
                const string &id(item.getID());
                const int cur_val = menu_selections.get(id);
                if (cur_val < item.getMinValue() || cur_val >= item.getMinValue()
                + item.getNumValues()) {
                    menu_selections.set(id, item.getMinValue());
                    changed = true;
                } else {
                    const MenuValue &val(menu.getItem(i).getValue(menu_selections.get(id)));
                    if (val.applyConstraints(menu_selections)) {
                        changed = true;
                    }
                }
            }
            if (!changed) break;
        }
    }

    void TransferMenuToWidgets(const Menu &menu, const MenuSelections &menu_selections,
                               map<string, MenuWidgets> &menu_widgets_map)
    {
        for (int i = 0; i < menu.getNumItems(); ++i) {
            const MenuItem &item = menu.getItem(i);
            const string &id = item.getID();
            MenuWidgets &mw = menu_widgets_map.find(id)->second;

            mw.list_model->elts.clear();
            mw.list_model->vals.clear();
            for (int j = item.getMinValue(); j < item.getMinValue() + item.getNumValues(); ++j) {
                if (menu.isSettingAllowed(menu_selections, id, j)) {
                    mw.list_model->elts.push_back(MenuValueToString(j, item.getValue(j)));
                    mw.list_model->vals.push_back(j);
                }
            }
            
            const int val_selected = menu_selections.get(id);
            for (int idx = 0; idx < mw.list_model->vals.size(); ++idx) {
                const int val_at_idx = mw.list_model->vals[idx];
                if (val_selected == val_at_idx) {
                    mw.dropdown->setSelected(idx);
                    break;
                }
            }
        }
    }

    int GetDropdownWidth(const gcn::Font &font, const MenuItem &item)
    {
        int widest_text = 0;
        for (int i = item.getMinValue(); i < item.getMinValue() + item.getNumValues(); ++i) {
            const string & text = item.getValue(i).getText();
            const int text_width = font.getWidth(text);
            if (text_width > widest_text) widest_text = text_width;
        }
        return widest_text + 4;
    }
}

    
class MenuScreenImpl : public gcn::ActionListener {
public:
    MenuScreenImpl(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &gui);
    void action(const gcn::ActionEvent &event);

    KnightsGame &knights_game;
    shared_ptr<Coercri::Window> window;

    scoped_ptr<GuiCentre> centre;
    scoped_ptr<GuiPanel> panel;
    scoped_ptr<gcn::Container> container;
    scoped_ptr<GuiButton> start_button, exit_button;
    scoped_ptr<gcn::Label> label, title;
    map<string, MenuWidgets> menu_widgets_map;
};

MenuScreenImpl::MenuScreenImpl(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &gui)
    : knights_game(kg), window(win)
{
    // will always have an Exit button
    exit_button.reset(new GuiButton("Exit"));
    exit_button->addActionListener(this);

    // create container
    container.reset(new gcn::Container);
    container->setOpaque(false);
    
    if (kg.server) {
        const int pad = 3, vpad_pretitle = 6, vpad_posttitle = 15, vpad_mid = 4, vpad_bot = 4, vspace = 2, extra_vspace = 10;
    
        // add the menu widgets to "container"
        const Menu &menu = kg.server_config->getMenu();
        int max_label_w = 0, max_dropdown_w = 0;
        for (int i = 0; i < menu.getNumItems(); ++i) {
            const MenuItem &item(menu.getItem(i));
            MenuWidgets menu_widgets(item, this);
            menu_widgets_map.insert(make_pair(item.getID(), menu_widgets));
            max_label_w = std::max(max_label_w, menu_widgets.label->getWidth());
            max_dropdown_w = std::max(max_dropdown_w, GetDropdownWidth(*menu_widgets.dropdown->getFont(), item));
        }
        int y = vpad_pretitle;

        const int container_w = pad + max_label_w + pad + max_dropdown_w + pad;
        title.reset(new gcn::Label("QUEST SELECTION"));
        container->add(title.get(), container_w/2 - title->getWidth()/2, y);
        y += title->getHeight() + vpad_posttitle;
        
        for (int i = 0; i < menu.getNumItems(); ++i) {
            const MenuItem &item(menu.getItem(i));
            map<string,MenuWidgets>::const_iterator mw = menu_widgets_map.find(item.getID());
            mw->second.dropdown->setWidth(max_dropdown_w);
            container->add(mw->second.label.get(), pad, y+1);
            container->add(mw->second.dropdown.get(), pad + max_label_w + pad, y);
            y += std::max(mw->second.label->getHeight(), mw->second.dropdown->getHeight()) + vspace;
            if (item.getSpaceAfter()) y += extra_vspace;
        }

        y += vpad_mid;
        
        // add the start and exit buttons
        start_button.reset(new GuiButton("Start"));
        start_button->addActionListener(this);
        container->add(start_button.get(), pad, y);
        container->add(exit_button.get(), container_w - exit_button->getWidth() - pad, y);
        y += std::max(exit_button->getHeight(), start_button->getHeight());
        y += vpad_bot;
        
        // resize the container
        container->setSize(container_w, y);

    } else {
        const int pad = 3, vspace = 6;
        
        // Add label to container
        label.reset(new gcn::Label("Waiting for game to begin"));
        container->add(label.get(), pad, pad);
        
        // Add exit button
        container->add(exit_button.get(),
                       pad + label->getWidth()/2 - exit_button->getWidth()/2,
                       pad + label->getHeight() + vspace);

        container->setSize(2*pad + label->getWidth(), 2*pad + vspace + label->getHeight() + exit_button->getHeight());
    }
    
    panel.reset(new GuiPanel(container.get()));
    centre.reset(new GuiCentre(panel.get()));
    gui.setTop(centre.get());
}

void MenuScreenImpl::action(const gcn::ActionEvent &event)
{
    if (event.getSource() == exit_button.get()) {
        // Exit
        // drop back to the title screen
        auto_ptr<Screen> scrn(new TitleScreen);
        knights_game.requestScreenChange(scrn);
        return;
    }

    if (event.getSource() == start_button.get()) {
        // Start Game
        // Try to start the InGameScreen. This can sometimes fail
        // (if there are dungeon generation errors)
        bool ok_to_start = true;
        auto_ptr<Screen> igs;
        try {
            igs.reset(new InGameScreen(knights_game, g_menu_selections));
        } catch (const DungeonGenerationFailed&) {
            ok_to_start = false;
        }
        if (ok_to_start) {
            knights_game.requestScreenChange(igs);
        }
        return;
    }

    // Must have been one of the drop downs
    const Menu &menu = knights_game.server_config->getMenu();
    for (int i = 0; i < menu.getNumItems(); ++i) {
        const MenuItem &item = menu.getItem(i);
        const string &id = item.getID();
        const MenuWidgets &mw = menu_widgets_map.find(id)->second;
        const gcn::DropDown *dropdown = mw.dropdown.get();
        if (dropdown == event.getSource()) {
            // Find out what was selected
            const int selected = dropdown->getSelected();
            if (selected >= 0 && selected < mw.list_model->vals.size()) {
                const int val = mw.list_model->vals[selected];
                
                // If they've changed a non-quest item then change the quest to CUSTOM
                if (id != "quest" && g_menu_selections.get(id) != val) {
                    g_menu_selections.set("quest", 0);
                }
                
                // Set the new value for the item and apply constraints
                g_menu_selections.set(id, val);
                ApplyMenuConstraints(menu, g_menu_selections);
                
                // Transfer back to widgets & redraw.
                TransferMenuToWidgets(menu, g_menu_selections, menu_widgets_map);
                window->invalidateAll();
                
                return;
            }
        }
    }

    // If we get here then the event came from an unknown widget. Just ignore it.
}

bool MenuScreen::start(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &gui)
{
    pimpl.reset(new MenuScreenImpl(kg, win, gui));

    if (kg.server) {
        if (g_first_time) {
            // Set up "g_menu_selections" if this is the first run through.
            
            for (int i = 0; i < kg.server_config->getMenu().getNumItems(); ++i) {
                const MenuItem &item(kg.server_config->getMenu().getItem(i));
                
                // Items are set to their min_value initially, except for "quest",
                // which is set to 1 (the first "real" quest value -- not CUSTOM).
                if (item.getID() == "quest") {
                    g_menu_selections.set("quest", 1);
                } else {
                    g_menu_selections.set(item.getID(), item.getMinValue());
                }
            }
            
            g_first_time = false;
        }

        // Ensure the constraints are applied properly...
        const Menu & menu(kg.server_config->getMenu());
        ApplyMenuConstraints(menu, g_menu_selections);
        TransferMenuToWidgets(menu, g_menu_selections, pimpl->menu_widgets_map);
        win->invalidateAll();
    }
    
    return true;
}


void MenuScreen::update()
{
    if (!pimpl->knights_game.server && pimpl->knights_game.remote) {
        if (pimpl->knights_game.remote->inAvail()) {
            // When we receive the 'M', we can leave the menu.
            // Anything else is an error.
            char c = pimpl->knights_game.remote->getUbyte();
            if (c == 'M') {
                auto_ptr<Screen> gscrn(new InGameScreen(pimpl->knights_game, g_menu_selections));
                pimpl->knights_game.requestScreenChange(gscrn);
            } else {
                throw UnexpectedError("network protocol error");
            }
        }
    }

    // Also check if the stream was closed -- if so, game has been abandoned and we should exit back to title screen.
    if (pimpl->knights_game.remote) {
        if (pimpl->knights_game.remote->getState() != Coercri::NetworkConnection::CONNECTED) {
            // stream was closed. we must drop back to the title screen
            auto_ptr<Screen> scrn(new TitleScreen);
            pimpl->knights_game.requestScreenChange(scrn);
        }
    }
}
