/*
 * potion_renderer.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "client_config.hpp"
#include "colour_change.hpp"
#include "gfx_manager.hpp"
#include "potion_renderer.hpp"
#include "round.hpp"
#include "task_manager.hpp"

PotionRenderer::PotionRenderer()
    : def_set(false), def_col(Coercri::Color(0,0,0,0))
{ }

void PotionRenderer::addGraphic(const Coercri::Graphic *g)
{
    gfx.push_back(g);
}

void PotionRenderer::addColour(Coercri::Color col)
{
    if (!def_set) {
        // First colour gets set as "default".
        def_set = true;
        def_col = col;
    } else {
        // Second and subsequent colours get added as ColourChanges.
        ColourChange cc;
        cc.add(def_col, col);
        colours.push_back(cc);
    }

}

void PotionRenderer::draw(int gvt, Coercri::GfxContext &gc, GfxManager &gm, int health, PotionMagic magic,
                          bool poison_immunity, int left, int top, float scale) const
{
    // Find out which graphic to draw
    if (health < 0) health = 0;
    if (health >= gfx.size()) health = gfx.size() - 1;
    const Coercri::Graphic *gr = gfx[health];

    const int width = Round(float(gr->getWidth()) * scale);
    const int height = Round(float(gr->getHeight()) * scale);
    
    // Check if poison immunity colour change should be used.
    if (poison_immunity) {
        const int s = (gvt / CfgInt("pot_flash_time")) % CfgInt("pi_flash_delay");
        if (s == 0) {
            // poison immunity: Use a white colour change
            ColourChange cc;
            cc.add(def_col, Coercri::Color(255, 255, 255));
            gm.drawTransformedGraphic(gc, left, top, *gr, width, height, cc);
            return;
        }
    }

    // Otherwise, work out colour to use based on "magic" value.
    if (magic < NO_POTION || magic > SUPER) return;  // magic out of range
    
    if (magic == NO_POTION) {
        // no potion: draw potion w/o colour change
        gm.drawTransformedGraphic(gc, left, top, *gr, width, height);
    } else if (magic == SUPER) {
        // super: use a random colour change
        // Use gvt as a random seed (we should not disturb g_rng during rendering).
        int xx = (gvt/CfgInt("pot_flash_time"))*1103515245;
        int r = (xx >> 16) & 0xff;
        int g = (xx >> 8) & 0xff;
        int b = (xx & 0xff);
        ColourChange cc;
        cc.add(def_col, Coercri::Color(r, g, b));
        gm.drawTransformedGraphic(gc, left, top, *gr, width, height, cc);
    } else {
        // draw with colour change appropriate to current magic
        if (magic-1 < 0 || magic-1 >= colours.size()) return;
        gm.drawTransformedGraphic(gc, left, top, *gr, width, height, colours[int(magic-1)]);
    }
}

void PotionRenderer::serialize(Coercri::NetworkConnection &remote, const GfxManager &gm) const
{
    remote.putUbyte(def_col.r);
    remote.putUbyte(def_col.g);
    remote.putUbyte(def_col.b);
    remote.putUbyte(def_col.a);
    remote.putLong(gfx.size());
    for (int i=0; i<gfx.size(); ++i) {
        remote.putLong(gm.getGraphicID(*gfx[i]));
    }
    remote.putLong(colours.size());
    for (int i=0; i<colours.size(); ++i) {
        colours[i].serialize(remote);
    }
}

PotionRenderer::PotionRenderer(Coercri::NetworkConnection &remote, ClientConfig &conf)
{
    def_col.r = remote.getUbyte();
    def_col.g = remote.getUbyte();
    def_col.b = remote.getUbyte();
    def_col.a = remote.getUbyte();
    int sz = remote.getLong();
    gfx.reserve(sz);
    for (int i=0; i<sz; ++i) {
        gfx.push_back(conf.getGraphic(remote.getLong()));
    }
    sz = remote.getLong();
    colours.reserve(sz);
    for (int i=0; i<sz; ++i) {
        colours.push_back(ColourChange(remote));
    }
}
